package audio;

import jkernel.FastCopy;

public final class AudioBlock implements FastCopy{

	private byte data[];
	private int desiredLen;

	protected int wpHeader;	// RNI pointer to header (message thread uses)
	private int wpData;		// RNI pointer to data

	// mark and notify stuff
	private boolean used, finished;
	private transient Object uselock, finishlock;

	private AudioBlock() {
		System.out.println("Creating audio block first");
		wpHeader = wpData = desiredLen = 0;
		data = null;
		used = finished = false;
		finishlock = new Object();
		uselock = new Object();
	}

	public AudioBlock(byte data[]) throws AudioException { // this is how you do output
		this();
		System.out.println("Creating audio block length "+data.length);
		if (data == null || data.length == 0) throw new AudioException("AudioBlock must be non-null and positive length");
		this.data = data;
		System.out.println("Created audio block length "+this.data.length);
	}
	public AudioBlock(int desiredLen) throws AudioException { // this is how you do input
		this();
		if (desiredLen <= 0) throw new AudioException("AudioBlock must have a positive length");
		this.desiredLen = desiredLen;
	}
	public Object clone() {
		AudioBlock b = new AudioBlock();
		b.data = data;
		b.desiredLen = desiredLen;
		// this makes too many copies, but too bad...
		return b;
	}
	public byte[] getData() {
		return data;
	}

	// a thread may call this to see when the block is done (in or output)
	public void waitForFinished() throws InterruptedException {
		if( finishlock == null)
			finishlock = new Object();
		synchronized(finishlock) {
			if (finished) return;
			finishlock.wait();
		}
	}

	protected void markAsFinished() {
		if( finishlock == null)
			finishlock = new Object();
		synchronized(finishlock) {
			if (data == null) {
				// copy the data from wpData
				data = new byte[desiredLen];
				copyDataN();
			}
			freeN(); // free the dll buffers
			finished = true;
			finishlock.notifyAll();
		}
	}

	protected boolean makeUseOut() {	// grabs the 'used' lock, permanently
		if( uselock == null)
			uselock = new Object();
		synchronized (uselock) {
			if (used || data == null) return false;
			used = true;
			return true;
		}
	}
	protected boolean makeUseIn() {	// grabs the 'used' lock, permanently
		if( uselock == null)
			uselock = new Object();
		synchronized (uselock) {
			if (used || data != null) return false;
			used = true;
			return true;
		}
	}
	private native void freeN();
	private native void copyDataN();

}
