
#include <windows.h>
#include <mmsystem.h>
//#include <winbase.h>
#include <MMREG.H>


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

__declspec(dllexport) DWORD __cdecl RNIGetCompatibleVersion();
#include <native.h>  // RNI declarations (part of Microsoft SDK for Java)
__declspec(dllexport) DWORD __cdecl RNIGetCompatibleVersion() { return RNIVER; }

#include "audio_Audio.h"
#include "audio_AudioIn.h"
#include "audio_AudioOut.h"
#include "audio_AudioBlock.h"
#include "audio_AudioMessage.h"
#include "audio_AudioMessageThread.h"

int gDebug;
#define dprintf gDebug && printf

#define CHECK_ERR(dev, err, msg) ((dev)->lastError = (err) ? (formatError((dev),(err),(msg)), 1): 0)
void formatError(struct Haudio_Audio *pThis, int err, char *msg);

//void CALLBACK waveOutHandler(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);
//void CALLBACK waveInHandler(HWAVEIN hwi, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);

#define RATE 11025
#define SAMPLEBITS 8

/// Audio.java
__declspec(dllexport)
long audio_Audio_getNumDevicesN(struct Haudio_Audio *pThis) // this should be static!!
{
	int nIN = waveInGetNumDevs();
	int nOUT = waveOutGetNumDevs();

	dprintf(" >>> There are %d input and %d output devices!\n", nIN,nOUT);
	return (nIN < nOUT)?nIN:nOUT;
}

__declspec(dllexport)
void checkCaps(int devID)
{
	WAVEOUTCAPS woc;

	waveOutGetDevCaps(devID,&woc,sizeof(WAVEOUTCAPS));
	printf("Wave Out Caps:\n");
	printf("  wMid : %d\n  wPid : %d\n  vDriverVersion : %d\n  Pname : %s\n  formats : %d\n  chans : %d\n",
		woc.wMid,woc.wPid,woc.vDriverVersion,woc.szPname,woc.dwFormats,woc.wChannels);
	if (WAVE_FORMAT_1M08 & woc.dwFormats) printf("  has 11.025 kHz, mono, 8-bit \n");
	if (WAVE_FORMAT_1M16 & woc.dwFormats) printf("  has 11.025 kHz, mono, 16-bit \n");
	if (WAVE_FORMAT_1S08 & woc.dwFormats) printf("  has 11.025 kHz, stereo, 8-bit \n");
	if (WAVE_FORMAT_1S16 & woc.dwFormats) printf("  has 11.025 kHz, stereo, 16-bit \n");
	if (WAVE_FORMAT_2M08 & woc.dwFormats) printf("  has 22.05 kHz, mono, 8-bit \n");
	if (WAVE_FORMAT_2M16 & woc.dwFormats) printf("  has 22.05 kHz, mono, 16-bit \n");
	if (WAVE_FORMAT_2S08 & woc.dwFormats) printf("  has 22.05 kHz, stereo, 8-bit \n");
	if (WAVE_FORMAT_2S16 & woc.dwFormats) printf("  has 22.05 kHz, stereo, 16-bit \n");
	if (WAVE_FORMAT_4M08 & woc.dwFormats) printf("  has 44.1 kHz, mono, 8-bit \n");
	if (WAVE_FORMAT_4M16 & woc.dwFormats) printf("  has 44.1 kHz, mono, 16-bit \n");
	if (WAVE_FORMAT_4S08 & woc.dwFormats) printf("  has 44.1 kHz, stereo, 8-bit \n");
	if (WAVE_FORMAT_4S16 & woc.dwFormats) printf("  has 44.1 kHz, stereo, 16-bit \n");

}

__declspec(dllexport)
void audio_Audio_initN(struct Haudio_Audio *pThis, long msgThreadID, long debug) // open the device and both channels
{
	HWAVEOUT hwo;
	HWAVEIN hwi;
	PCMWAVEFORMAT wfx;
	int err;

	gDebug = debug;

	wfx.wf.wFormatTag = WAVE_FORMAT_PCM;
	wfx.wf.nChannels = 1;
	wfx.wf.nSamplesPerSec = RATE;
	wfx.wf.nAvgBytesPerSec = RATE;
	wfx.wf.nBlockAlign = 1;
	wfx.wBitsPerSample = SAMPLEBITS;

//	printf(" >>> device open (this = 0x%x, thread = %d)\n", pThis, msgThreadID);

//	checkCaps(pThis->devID);

	dprintf(" >>> opening device %d output\n", pThis->devID);
	err = waveOutOpen(&hwo,/*WAVE_MAPPER*/pThis->devID,(WAVEFORMATEX *)&wfx,(long)msgThreadID,0,
		CALLBACK_THREAD );
	if (CHECK_ERR(pThis,err,"Opening Device Output")) return;
	err = waveOutSetVolume(hwo, 0xFFFFFFFF);
	if (CHECK_ERR(pThis,err,"Setting Output Volume")) return;

	dprintf(" >>> opening device %d input\n", pThis->devID);
	err = waveInOpen(&hwi,/*WAVE_MAPPER*/pThis->devID,(WAVEFORMATEX *)&wfx,(long)msgThreadID,0,
		CALLBACK_THREAD );
	if (CHECK_ERR(pThis,err,"Opening Device Input")) {
		waveOutClose(hwo);
		return;
	}

	pThis->hwo = (long)hwo;
	pThis->hwi = (long)hwi;
}

__declspec(dllexport)
void audio_Audio_closeN(struct Haudio_Audio *pThis)
{
	int err1, err2, err3, err4;
	dprintf("Closing device (out = %d, in = %d)\n", (HWAVEOUT)pThis->hwo, (HWAVEOUT)pThis->hwi);
	err1 = waveOutReset((HWAVEOUT)pThis->hwo);
	err2 = waveInReset((HWAVEIN)pThis->hwi);
	err3 = waveOutClose((HWAVEOUT)pThis->hwo);
	err4 = waveInClose((HWAVEIN)pThis->hwi);
	if (CHECK_ERR(pThis,err1,"Closing Device Output ")) return;
	if (CHECK_ERR(pThis,err2,"Closing Device Output ")) return;
	if (CHECK_ERR(pThis,err3,"Closing Device Output ")) return;	
	CHECK_ERR(pThis, err4, "Closing Device Input");
}

// AudioMessageThread.java
__declspec(dllexport)
long audio_AudioMessageThread_getThreadIDN(struct Haudio_AudioMessageThread *pThis)
{
	return GetCurrentThreadId();
}

__declspec(dllexport)
long audio_AudioMessageThread_getMessageN(struct Haudio_AudioMessageThread *pThis, struct Haudio_AudioMessage *pMsg)
{
	MSG msg;
	UINT min = WOM_OPEN, max = WIM_DATA;	// these bound the messages that we want
	int status;
	// we will be down here a long time (nearly all the time) so we want to enable GC
	struct { Haudio_AudioMessage *safe; } gcsafe;
	GCFrame gcf;
	GCFramePush(&gcf, &gcsafe, sizeof(gcsafe));
	gcsafe.safe = pMsg;
	GCEnable();

//	printf(" +++ Thread %d Waiting for message\n", GetCurrentThreadId());
	status = GetMessage(&msg, (HWND)NULL, min, max);
//	printf(" +++ Thread %d Got message: %d %d %d\n", GetCurrentThreadId(), msg.message,msg.wParam,msg.lParam);

	GCDisable();
	GCFramePop(&gcf);

	pMsg = gcsafe.safe;

	if (status < 0) {
		msg.message = status;
	}
	pMsg->message = msg.message;
	pMsg->param1 = msg.wParam;
	pMsg->param2 = msg.lParam;

	return status;
}

/// AudioOut.java
__declspec(dllexport)
void audio_AudioOut_startN(struct Haudio_AudioOut *pThis) {
	int err = waveOutRestart((HWAVEOUT)pThis->dev->hwo);
	CHECK_ERR(pThis->dev,err, "Starting Output");
}

__declspec(dllexport)
void audio_AudioOut_stopN(struct Haudio_AudioOut *pThis) {
	int err = waveOutPause((HWAVEOUT)pThis->dev->hwo);
	CHECK_ERR(pThis->dev,err, "Stopping Output");
}

__declspec(dllexport)
void audio_AudioOut_putBlockN(struct Haudio_AudioOut *pThis, struct Haudio_AudioBlock *pBlock) {
	int err;
	HWAVEOUT hwo = (HWAVEOUT)pThis->dev->hwo;
	int dlen = pBlock->data->length;
	char *data = (char *)malloc(dlen);
	int whs = sizeof(WAVEHDR);
	PWAVEHDR pwh = (PWAVEHDR)malloc(whs);

	memcpy(data, pBlock->data->body, dlen); // we NEED a copy because of gc... no way to pin memory

	pBlock->wpHeader = (int)pwh;
	pBlock->wpData = (int)data;

	pwh->lpData = data;
	pwh->dwBufferLength = dlen;
	pwh->dwLoops = 0;
	pwh->dwFlags = 0;

    dprintf(" >>> buffer length %d for output\n", pwh->dwBufferLength);

	err = waveOutPrepareHeader(hwo,pwh,whs);
	if (CHECK_ERR(pThis->dev,err,"Preparing Output Header")) {
		audio_AudioBlock_freeN(pBlock);
		return;
	}

	err = waveOutWrite(hwo, pwh, whs);
	if (CHECK_ERR(pThis->dev,err,"Sending Output Buffer")) {
		audio_AudioBlock_freeN(pBlock);
		return;
	}
}

// AudioIn.java

__declspec(dllexport)
void audio_AudioIn_startN(Haudio_AudioIn *pThis) {
	int err = waveInStart((HWAVEIN)pThis->dev->hwi);
	CHECK_ERR(pThis->dev,err, "Starting Input");
}
__declspec(dllexport)
void audio_AudioIn_stopN(Haudio_AudioIn *pThis) {
	int err = waveInStop((HWAVEIN)pThis->dev->hwi);
	CHECK_ERR(pThis->dev,err, "Stopping Input");
}

__declspec(dllexport)
void audio_AudioIn_addBlockN(struct Haudio_AudioIn *pThis, struct Haudio_AudioBlock *pBlock)
{
	int err;
	HWAVEIN hwi = (HWAVEIN)pThis->dev->hwi;
	int whs = sizeof(WAVEHDR);
	PWAVEHDR pwh = (PWAVEHDR)malloc(whs);
	int dlen = pBlock->desiredLen;
	char *data = (char *)malloc(dlen);

	pBlock->wpHeader = (int)pwh;
	pBlock->wpData = (int)data;

	pwh->lpData = data;
	pwh->dwBufferLength = dlen;
	pwh->dwLoops = 0;
	pwh->dwFlags = 0;

//    printf(" >>> buffer length %d\n", pwh->dwBufferLength);

	err = waveInPrepareHeader(hwi,pwh,whs);
	if (CHECK_ERR(pThis->dev,err,"Preparing Input Header")) {
		audio_AudioBlock_freeN(pBlock);
		return;
	}

	err = waveInAddBuffer(hwi, pwh, whs);
	if (CHECK_ERR(pThis->dev,err,"Sending Input Buffer")) {
		audio_AudioBlock_freeN(pBlock);
		return;
	}
}


// AudioBlock.java
__declspec(dllexport)
void audio_AudioBlock_freeN(struct Haudio_AudioBlock *pThis)
{
	if (pThis->wpHeader) free((char *)pThis->wpHeader);
	if (pThis->wpData) free((char *)pThis->wpData);
}
__declspec(dllexport)
void audio_AudioBlock_copyDataN(struct Haudio_AudioBlock *pThis)
{
	if (pThis->wpData && pThis->data) {
		memcpy(pThis->data->body, (char *)pThis->wpData, pThis->desiredLen);
	}
}

///// error interpreting
void formatError(struct Haudio_Audio *pThis, int err, char *msg)
{
	char buf[256];

	dprintf(" >>> error comming\n");
	switch(err) {
	case WAVERR_BADFORMAT    :
		sprintf(buf,"%s: unsupported wave format\n", msg);
		break;
	case WAVERR_STILLPLAYING :
		sprintf(buf,"%s: still something playing\n", msg);
		break;
	case WAVERR_UNPREPARED   :
		sprintf(buf,"%s: header not prepared\n", msg);
		break;
	case WAVERR_SYNC         :
		sprintf(buf,"%s: device is synchronous\n", msg);
		break;
		
	case MMSYSERR_ERROR        :
		sprintf(buf,"%s: unspecified error\n", msg);
		break;
	case MMSYSERR_BADDEVICEID:
		sprintf(buf,"%s: device ID out of range\n", msg);
		break;
	case MMSYSERR_NOTENABLED :
		sprintf(buf,"%s: driver failed enable\n", msg);
		break;
	case MMSYSERR_ALLOCATED  :
		sprintf(buf,"%s: device already allocated\n", msg);
		break;
	case MMSYSERR_INVALHANDLE:
		sprintf(buf,"%s: device handle is invalid\n", msg);
		break;
	case MMSYSERR_NODRIVER   :
		sprintf(buf,"%s: no device driver present\n", msg);
		break;
	case MMSYSERR_NOMEM      :
		sprintf(buf,"%s: memory allocation error\n", msg);
		break;
	case MMSYSERR_NOTSUPPORTED:
		sprintf(buf,"%s: function isn't supported\n", msg);
		break;
	case MMSYSERR_BADERRNUM  :
		sprintf(buf,"%s: error value out of range\n", msg);
		break;
	case MMSYSERR_INVALFLAG  :
		sprintf(buf,"%s: invalid flag passed\n", msg);
		break;
	case MMSYSERR_INVALPARAM :
		sprintf(buf,"%s: invalid parameter passed\n", msg);
		break;
	case MMSYSERR_HANDLEBUSY :
		sprintf(buf,"%s: handle being used\n", msg);
		break;
	case MMSYSERR_INVALIDALIAS:
		sprintf(buf,"%s: specified alias not found\n", msg);
		break;
	case MMSYSERR_BADDB      :
		sprintf(buf,"%s: bad registry database\n", msg);
		break;
	case MMSYSERR_KEYNOTFOUND:
		sprintf(buf,"%s: registry key not found\n", msg);
		break;
	case MMSYSERR_READERROR  :
		sprintf(buf,"%s: registry read error\n", msg);
		break;
	case MMSYSERR_WRITEERROR :
		sprintf(buf,"%s: registry write error\n", msg);
		break;
	case MMSYSERR_DELETEERROR:
		sprintf(buf,"%s: registry delete error\n", msg);
		break;
	case MMSYSERR_VALNOTFOUND:
		sprintf(buf,"%s: registry value not found\n", msg);
		break;
	case MMSYSERR_NODRIVERCB :
		sprintf(buf,"%s: driver does not call DriverCallback\n", msg);
		break;
	case MMSYSERR_NOERROR:
		return;
	default:
		sprintf(buf,"%s: unknown status %d(0x%x)\n", msg,err,err);
		break;
	}

	dprintf(" >>> there was an error: %s\n", buf);
	Field_SetObject((HObject *)pThis,
					Class_GetField(Object_GetClass((HObject *)pThis),"lastErrorMsg"), 
					(HObject *)makeJavaString(buf, strlen(buf)));
}
