package audio;

import java.io.*;
import java.util.*;

public class Samples {
	public static final AudioBlock DONG = createTestSound(40,150,9000);

	static AudioBlock createTestSound(int p1,int p2,int decay)
	{
		int len = 11000*3;
		byte data[] = new byte[len];
		float pi = (float)3.1415;
		for(int i = 0; i < len; i++) {
			float j = (float)i;
			data[i] = (byte)(127 + (byte)(127*(Math.sin(j*2*pi/p1)*Math.sin(j*2*pi/p2)*Math.exp(-j/decay))));
		}
		try {
			return new AudioBlock(data);
		} catch (AudioException e) {
			return null;
		}
	}

		/// nice utilities ///////////////
	
	private static int getInt(InputStream in) throws IOException {
		int i;
		i = in.read();
		i |= in.read() << 8;
		i |= in.read() << 16;
		i |= in.read() << 24;
		return i;
	}
	private static char getChar(InputStream in) throws IOException {
		int i;
		i = in.read();
		i |= in.read() << 8;
		return (char)i;
	}

	private static int chunkLen(DataInputStream in, String name) throws IOException, AudioException
	{
		byte hdr[] = new byte[4];
		if (in.read(hdr) != 4) throw new IOException("Could not read file");
		if (!(new String(hdr).equals(name))) throw new AudioException("bad file format");
		int len = getInt(in);
		return len;
	}

	public static AudioBlock[] readWavFile(String name) throws IOException, AudioException
	{
		DataInputStream in = new DataInputStream(new FileInputStream(name));
		// read the RIFF block
		chunkLen(in,"RIFF");	// this is the file size
		byte wave[] = new byte[4];
		if (in.read(wave) != 4) throw new IOException("Could not read file");
		if (!(new String(wave).equals("WAVE"))) throw new AudioException("bad file format");

		int fmtlen = chunkLen(in,"fmt "); // cant figure out what this means
		char pcm = getChar(in);
		char chans = getChar(in);
		int rate = getInt(in);
		int bytesPerSec = getInt(in);
		char bytesPerSamp = getChar(in);
		char bitsPerSamp = getChar(in);

		if (pcm != 1) throw new AudioException("cant only do PCM");
		if (chans != 1) throw new AudioException("cant do multiple channels");
		if (rate != 11025) throw new AudioException("can only do 11025 hertz");
		if (bytesPerSamp != 1) throw new AudioException("can only do 1 byte/sample");
		if (bitsPerSamp != 8) throw new AudioException("can only do 8 bit samples");
		if (bytesPerSec < 0) throw new AudioException("can only send positive bitrate");

		if (fmtlen == 18) {
			// we have to skip some
			in.skip(14);
		}
		int len = chunkLen(in,"data");
		int numBlocks = (len+11024) / 11025;
		AudioBlock blocks[] = new AudioBlock[numBlocks];
		for(int i = 0; i < numBlocks; i++) {
			int l = len > 11025 ? 11025 : len;
			byte data[] = new byte[l];
			len -= in.read(data);
			blocks[i] = new AudioBlock(data);
		}
		if (len != 0) throw new AudioException("didnt read all data");
		in.close();
		return blocks;
	}

	private static void writeString(OutputStream out, String s) throws IOException {
		for(int i = 0; i < s.length(); i++) {
			out.write(s.charAt(i));
		}
	}
	private static void writeInt(OutputStream out, int i) throws IOException {
		out.write(i & 0xFF); i = i >> 8;
		out.write(i & 0xFF); i = i >> 8;
		out.write(i & 0xFF); i = i >> 8;
		out.write(i & 0xFF); i = i >> 8;
	}
	private static void writeChar(OutputStream out, int i) throws IOException {
		out.write(i & 0xFF); i = i >> 8;
		out.write(i & 0xFF); i = i >> 8;
	}

	public static void writeWavFile(String name, AudioBlock blocks[]) throws IOException
	{
		OutputStream out = new FileOutputStream(name);
		
		// how much data?
		int len = 0;
		for(int i = 0; i < blocks.length; i++) {
			len += blocks[i].getData().length;
		}
		int filelen = len + 36;

		writeString(out,"RIFF");
		writeInt(out,filelen);
		writeString(out,"WAVE");

		writeString(out,"fmt ");
		writeInt(out, 16);
		writeChar(out, 1);	// PCM
		writeChar(out, 1);	// mono
		//writeInt(out, 11025); // rate
		//writeInt(out, 11025); // bytes per sec
		writeInt(out, 8000); // rate
		writeInt(out, 8000); // bytes per sec
		writeChar(out, 1); // 1 byte per sample
		writeChar(out, 8); // 8 bits per sample

		writeString(out, "data");
		writeInt(out, len);
		for(int i = 0; i < blocks.length; i++) {
			out.write(blocks[i].getData());
		}
		out.close();
	}
}

