package audio;

import java.io.*;
import java.util.*;

class Test {
	public static final void info(String msg) {
		System.out.println(msg);
	}
	
	static byte testSound[];
	static {
		int len = 11000*1;
		testSound = new byte[len];
		int p1 = 40, p2 = 150, decay = 9000;
		float pi = (float)3.1415;
		for(int i = 0; i < len; i++) {
			float j = (float)i;
			testSound[i] = (byte)(127 + (byte)(127*(Math.sin(j*2*pi/p1)*Math.sin(j*2*pi/p2)*Math.exp(-j/decay))));
		//	System.out.print((short)(testSound[i])+"\n");
		}
	}

	public static final String syntax =
		"Syntax:  audio.Test [-d dev] [-f wavfile]";

	public static void main(String args[]) {
		try {
			int device = 0;
			String name = null;
			try {
				for(int i = 0; i < args.length; i++) {
					if (args[i].equals("-d")) {
							i++;
							device = Integer.parseInt(args[i]);
					} else if (args[i].equals("-f")) {
						i++;
						name = args[i];
					} else throw new Exception();
				}
			} catch (Exception e) {
				System.out.println(syntax);
				return;				
			}
			int n = Audio.getNumDevices();
			info("There are "+n+" audio drivers");
			info("Opening device "+device);
			Audio dev = new Audio(device);
			info("Device is named \"" + dev.devName+"\"");
			if (name != null) {
				AudioBlock blocks[] = Samples.readWavFile(name);
				for(int i = 0; i < blocks.length; i++) {
					dev.out.putBlock(blocks[i]);
				}
				return;
			}

			AudioBlock b = null;
			int choice = 0;
			BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 
			while(choice != 9) {
				System.out.println(
					"1  Start output\n"+
					"2  Stop output\n"+
					"3  Start input\n"+
					"4  Stop input\n"+
					"5  Send test block\n"+
					"6  Send recd block\n"+
					"7  Add block to input queue\n"+
					"8  Save to file sample.wav\n"+
					"\n9  Exit\n");
				System.out.print("Choice: ");
				System.out.flush();
				try {
					choice = Integer.parseInt(in.readLine());
				} catch (Exception e) {
					choice = -1;
				}
				switch(choice) {
				case 1:
					dev.out.start();
					break;
				case 2:
					dev.out.stop();
					break;
				case 3:
					dev.in.start();
					break;
				case 4:
					dev.in.stop();
					break;
				case 5:
					System.out.println("Sending test sound");
					b = new AudioBlock(testSound);
					System.out.println("About to send");
					dev.out.putBlock(b);
					break;
				case 6:
					b.waitForFinished();
					break;
				case 7:
					b = new AudioBlock(11000*3);
					dev.in.addBlock(b);
					break;
				case 8:
					break;
				default:
					System.out.println("\nNot a valid choice...");
					break;
				}
			}
			dev.close();
			dev = null;
		} catch (Exception e) {
			e.printStackTrace(System.out);
		}
	}
}
